import os
import sys
import requests
import json
import logging
import csv
import traceback

import matplotlib.pyplot as plt
import matplotlib
import numpy as np

from PIL import Image,ImageDraw,ImageFont
from datetime import datetime


libdir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'lib')
if os.path.exists(libdir):
    sys.path.append(libdir)
from waveshare_epd import epd4in37g



logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
#logger.setLevel(logging.DEBUG)

install_id="123456"
username="useremail@domain.com"
password="securepassword"


resdir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'resources')
csv_file_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'latest.val')
#his_image_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'history.png')

#WHITE = (255,255,255)
#BLACK = (0,0,0)
#RED = (255,0,0)
#YELLOW = (255,255,0)

login_url = 'https://vrmapi.victronenergy.com/v2/auth/login'
diags_url = "https://vrmapi.victronenergy.com/v2/installations/"+install_id+"/diagnostics?count=1000"
batterysummary_url = "https://vrmapi.victronenergy.com/v2/installations/"+install_id+"/stats"

def get_token(username, password):
    login_string = '{"username":"'+username+'","password":"'+password+'"}'
    response = requests.post(login_url , login_string)
    return json.loads(response.text)["token"]

# Function not used.
# these values do not update frequently
def get_data_vrm_api(): 
    token = get_token(username, password)
    headers = {'X-Authorization': "Bearer " + token }
    response = requests.get(diags_url, headers=headers)

    data = response.json()["records"]

    batterySoC=[element['rawValue'] for element in data if element['code']=="SOC"][0]
    batterypower=[element['rawValue'] for element in data if element['code']=="bp"][0]

    solarpowerDC=[element['rawValue'] for element in data if element['code']=="PVP"][0]
    solarpowerAC=[element['rawValue'] for element in data if element['code']=="pP1"][0]
    solarpower=float(solarpowerDC)+float(solarpowerAC)

    loadpower=[element['rawValue'] for element in data if element['code']=="o1"][0]
    gridpower=[element['rawValue'] for element in data if element['code']=="IP1"][0]

    timestamp=(datetime.fromtimestamp([element['timestamp'] for element in data if element['code']=="SOC"][0]))

    print("SoC (%): ",batterySoC)
    print("Battery charge/discharge (W) :" ,batterypower)
    print("Solar power (W): ",solarpower)
    print("Load (W): ",loadpower)
    print("gridpower (W): ",gridpower)
    print("data taken at : ", timestamp)
    return [float(batterySoC), float(batterypower)/1000, float(solarpower)/1000, float(loadpower)/1000, float(gridpower)/1000, str(timestamp)]

def get_data():
    logger.info("reading local data...")
    with open(csv_file_path,'r') as file:
        csvreader = csv.reader(file)
        for row in csvreader:
            data = row
    data[0] = float(data[0])
    data[1] = float(data[1])/1000
    data[2] = float(data[2])/1000
    data[3] = float(data[3])/1000
    data[4] = float(data[4])/1000
    #print(data)
    return data

def fig2img(fig):
    """Convert a Matplotlib figure to a PIL Image and return it"""
    import io
    buf = io.BytesIO()
    fig.savefig(buf)
    buf.seek(0)
    img = Image.open(buf)
    return img

def get_graph_data():
    logger.info("retrieving historical data...")
    token = get_token(username, password)
    headers = {'X-Authorization': "Bearer " + token }
    response = requests.get(batterysummary_url, headers=headers)

    consumption_list = response.json()["records"]["total_consumption"]
    solar_list = response.json()["records"]["total_solar_yield"]
    soc_list = response.json()["records"]["bs"]


    logger.debug("TABLE:")
    data=[]
    con_out=[]
    sol_out=[]
    soc_out=[]

    # add values for no data
    for con in consumption_list:
        solval=find_value(solar_list,con[0])
        socval=find_value(soc_list,con[0])
        logger.debug(f"{con[1]:.2f} {solval:.2f} {socval:.2f}")
        con_out.append(con[1])
        sol_out.append(solval)
        soc_out.append(socval)

    return con_out,sol_out,soc_out

# return the value at a given timestamp
def find_value(data, target_value):
    for i, sublist in enumerate(data):
        if target_value in sublist:
            sublist_index = sublist.index(target_value)
            #print(f"Found {target_value} at index ({i}, {sublist_index})")
            return data[i][1]
    return 0


def graph_data(data):
    logger.info("graphing data...")

    my_dpi=72
    matplotlib.rcParams['text.antialiased'] = False
    matplotlib.rcParams['lines.antialiased'] = False
    matplotlib.rcParams['font.size'] = 16


    plt.figure(figsize=(368/my_dpi,200/my_dpi),dpi=my_dpi)

    bar_width = 0.5
    ind = np.arange(len(data[0]))

    #plot compumption and solar
    plt.bar(ind-bar_width/2,data[1],bar_width,color='yellow')
    plt.bar(ind+bar_width/2,data[0],bar_width,color='red')

    #hide the top spline
    ax=plt.gca()
    ax.spines['top'].set_visible(False)

    plt.twinx()

    #plot Battery
    plt.ylim(0,103)
    plt.plot(data[2],color="black")

    #hide the x axis numbers
    plt.xticks([])
    #plt.yticks([])

    #hide the top spine
    ax=plt.gca()
    ax.spines['top'].set_visible(False)


    #adjust the margins
    plt.subplots_adjust(left=0.10, right=0.88)

    #logger.info("writing to image...")
    #plt.savefig(his_image_path,dpi=my_dpi)
    #plt.show()
    return plt

# Fill the battery Bar Rectangle
def fillBatRec(percentage,x,y):
    height = int(percentage/100*39)
    draw.rectangle((x,y-height,x+19,y), fill=epd.BLACK)

def drawArrows(direction,number,x,y,canvas):
     #draw arrows
    arrow_image=Image.open(os.path.join(resdir, 'arrow.gif'))
    dx = 0
    dy = 0
    if direction == "LEFT":
        dx = 17
    elif direction == "RIGHT":
        arrow_image = arrow_image.transpose(Image.FLIP_LEFT_RIGHT)
        dx = 17
    elif direction == "DOWN":
        arrow_image = arrow_image.rotate(90, expand=True)
        dy = 17
    elif direction == "UP":
        arrow_image = arrow_image.rotate(270, expand=True)
        dy = 17
    
    for index in range(1, number+1):
        canvas.paste(arrow_image, (x,y))
        x += dx
        y += dy

try:
    epd = epd4in37g.EPD()
    epd.init()
    ##epd.Clear()
    
    print (" ")
    logger.info(datetime.now())
    
    font24 = ImageFont.truetype(os.path.join(resdir, 'Font.ttc'), 24)
    font18 = ImageFont.truetype(os.path.join(resdir, 'Font.ttc'), 18)
    font40 = ImageFont.truetype(os.path.join(resdir, 'Font.ttc'), 40)

    data = get_data()


    Vimage = Image.new('RGB', (368, 512), epd.WHITE)  # 255: clear the frame

    Vimage = Image.open(os.path.join(resdir, 'Background.png'))
    draw = ImageDraw.Draw(Vimage)



    #draw text
    draw.text((12, 190), "{:.0f}%".format(data[0]) , font = font40, fill = epd.BLACK) #battery %
    draw.text((197, 32), "{:.1f}kW".format(data[2]) , font = font40, fill = epd.BLACK) #solar power
    draw.text((197, 250), "{:.1f}kW".format(data[3]) , font = font40, fill = epd.BLACK) #load Power
    #small text
    draw.text((80, 132), "{:.1f}kW".format(data[1]) , font = font24, fill = epd.BLACK) #battery power
    draw.text((230, 132), "{:.1f}kW".format(data[4]) , font = font24, fill = epd.BLACK) #grid power
    #date updated
    draw.text((8, 5), data[5].split()[0] , font = font18, fill = epd.BLACK) #grid power
    draw.text((8, 25), data[5].split()[1] , font = font18, fill = epd.BLACK) #grid power

    #fill battery
    fillBatRec(data[0],33,172)

    y_pos = 170
    bat_x = 97
    # Battery Arrows
    if data[1] < -3.33:
        # > > >
        drawArrows("RIGHT", 3, bat_x,y_pos, Vimage)
    elif data[1] < -1.66:
        # > >
        drawArrows("RIGHT", 2, bat_x,y_pos, Vimage)
    elif data[1] < -0.1:
        # >
        drawArrows("RIGHT", 1, bat_x,y_pos, Vimage)
    elif data[1] < 0.1:
        # x
        drawArrows("RIGHT", 0, bat_x,y_pos, Vimage)
    elif data[1] < 1.66:
        # <
        drawArrows("LEFT", 1, bat_x,y_pos, Vimage)
    elif data[1] < 3.33:
        # < <
        drawArrows("LEFT", 2, bat_x,y_pos, Vimage)
    else:
        # < < <
        drawArrows("LEFT", 3, bat_x,y_pos, Vimage)

    grd_x = 239

    #Grid Arrows
    if data[4] < -3.33:
        # > > >
        drawArrows("RIGHT", 3, grd_x, y_pos, Vimage)
    elif data[4] < -1.66:
        # > >
        drawArrows("RIGHT", 2, grd_x, y_pos, Vimage)
    elif data[4] < -0.1:
        # >
        drawArrows("RIGHT", 1, grd_x, y_pos, Vimage)
    elif data[4] < 0.1:
        # x
        drawArrows("LEFT", 0, grd_x, y_pos, Vimage)
    elif data[4] < 3.33:
        # <
        drawArrows("LEFT", 1, grd_x, y_pos, Vimage)
    elif data[4] < 6.66:
        # < <
        drawArrows("LEFT", 2, grd_x, y_pos, Vimage)
    else:
        # < < <
        drawArrows("LEFT", 3, grd_x, y_pos, Vimage)


    #Load Arrows
    if data[3] > 0.1:
        # \/
        drawArrows("DOWN", 1, 157, 208, Vimage)

    #Solar Arrows
    if data[2] > 0.1:
        # \/
        drawArrows("DOWN", 1, 157, 100, Vimage)
    

    #Plot the graphing data
    
    his_image=fig2img(graph_data(get_graph_data()))
    #his_image=Image.open(his_image_path)
    Vimage.paste(his_image,(1,312))

    rotated = Vimage.transpose(Image.ROTATE_270) 

    logger.info("Writing to Display")
    epd.display(epd.getbuffer(rotated))
    #rotated.show()
    #Vimage.show()

except IOError as e:
    print(e)

except KeyboardInterrupt:    
    print("ctrl + c:")

finally:
    logger.info("Going to sleep")
    epd.sleep()
    #logger.info("exiting")
    #exit()





